"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Any, Dict
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from gluonts.dataset.field_names import FieldName
from gluonts.torch.util import copy_parameters
from gluonts.env import env
from gluonts.dataset.common import Dataset
from gluonts.torch.model.predictor import PyTorchPredictor
from gluonts.transform import SelectFields, Transformation
from gluonts.itertools import maybe_len
from gluonts.transform import (
    AddObservedValuesIndicator,
    CDFtoGaussianTransform,
    Chain,
    InstanceSampler,
    InstanceSplitter,
    RenameFields,
    ValidationSplitSampler,
    TestSplitSampler,
    cdf_to_gaussian_forward_transform,
)
from pts import Trainer
from pts.model import PyTorchEstimator
from pts.model.estimator import TrainOutput
from pts.model.utils import get_module_forward_input_names
from pts.dataset.loader import TransformedDataset, TransformedIterableDataset

from .network import (
    TACTiSPredictionNetwork,
    TACTiSTrainingNetwork,
    TACTiSPredictionNetworkInterpolation,
)

from tactis.gluon.metrics import compute_validation_metrics, SplitValidationTransform
from gluonts.torch.batchify import batchify


class SingleInstanceSampler(InstanceSampler):
    """
    Randomly pick a single valid window in the given time series.
    This fix the bias in ExpectedNumInstanceSampler which leads to varying sampling frequency
    of time series of unequal length, not only based on their length, but when they were sampled.
    """

    """End index of the history"""

    def __call__(self, ts: np.ndarray) -> np.ndarray:
        a, b = self._get_bounds(ts)
        window_size = b - a + 1

        if window_size <= 0:
            return np.array([], dtype=int)

        indices = np.random.randint(window_size, size=1)

        return indices + a


class TACTiSEstimator(PyTorchEstimator):
    """
    The compatibility layer between TACTiS and GluonTS / PyTorchTS.
    """

    def __init__(
        self,
        model_parameters: Dict[str, Any],
        num_series: int,
        history_length: int,
        prediction_length: int,
        freq: str,
        trainer: Trainer,
        cdf_normalization: bool = False,
        num_parallel_samples: int = 1,
    ):
        """
        A PytorchTS wrapper for TACTiS

        Parameters:
        -----------
        model_parameters: Dict[str, Any]
            The parameters that will be sent to the TACTiS model.
        num_series: int
            The number of series in the multivariate data.
        history_length: int
            How many time steps will be sent to the model as observed.
        prediction_length: int
            How many time steps will be sent to the model as unobserved, to be predicted.
        freq: str
            The frequency of the series to be forecasted.
        trainer: Trainer
            A Pytorch-TS trainer object
        cdf_normalization: bool, default to False
            If set to True, then the data will be transformed using an estimated CDF from the
            historical data points, followed by the inverse CDF of a Normal(0, 1) distribution.
            Should not be used concurrently with the standardization normalization option in TACTiS.
        num_parallel_samples: int, default to 1
            How many samples to draw at the same time during forecast.
        """
        super().__init__(trainer=trainer)

        self.model_parameters = model_parameters

        self.num_series = num_series
        self.history_length = history_length
        self.prediction_length = prediction_length
        self.freq = freq

        self.cdf_normalization = cdf_normalization
        self.num_parallel_samples = num_parallel_samples

    def create_training_network(self, device: torch.device) -> nn.Module:
        """
        Create the encapsulated TACTiS model which can be used for training.

        Parameters:
        -----------
        device: torch.device
            The device where the model parameters should be placed.

        Returns:
        --------
        model: nn.Module
            An instance of TACTiSTrainingNetwork.
        """
        return TACTiSTrainingNetwork(
            num_series=self.num_series,
            model_parameters=self.model_parameters,
        ).to(device=device)

    def create_instance_splitter(self, mode: str) -> Transformation:
        """
        Create and return the instance splitter needed for training, validation or testing.

        Parameters:
        -----------
        mode: str, "training", "validation", or "test"
            Whether to split the data for training, validation, or test (forecast)

        Returns
        -------
        Transformation
            The InstanceSplitter that will be applied entry-wise to datasets,
            at training, validation and inference time based on mode.
        """
        assert mode in ["training", "validation", "test"]

        if mode == "training":
            instance_sampler = SingleInstanceSampler(
                min_past=self.history_length,  # Will not pick incomplete sequences
                min_future=self.prediction_length,
            )
        elif mode == "validation":
            instance_sampler = ValidationSplitSampler(
                min_past=self.history_length,  # Will not pick incomplete sequences
                min_future=self.prediction_length,
            )
        elif mode == "test":
            # This splitter takes the last valid window from each multivariate series,
            # so any multi-window split must be done in the data definition.
            instance_sampler = TestSplitSampler()

        if self.cdf_normalization:
            normalize_transform = CDFtoGaussianTransform(
                cdf_suffix="_norm",
                target_field=FieldName.TARGET,
                target_dim=self.num_series,
                max_context_length=self.history_length,
                observed_values_field=FieldName.OBSERVED_VALUES,
            )
        else:
            normalize_transform = RenameFields(
                {
                    f"past_{FieldName.TARGET}": f"past_{FieldName.TARGET}_norm",
                    f"future_{FieldName.TARGET}": f"future_{FieldName.TARGET}_norm",
                }
            )

        instance_sampler = (
            InstanceSplitter(
                target_field=FieldName.TARGET,
                is_pad_field=FieldName.IS_PAD,
                start_field=FieldName.START,
                forecast_start_field=FieldName.FORECAST_START,
                instance_sampler=instance_sampler,
                past_length=self.history_length,
                future_length=self.prediction_length,
                time_series_fields=[FieldName.OBSERVED_VALUES],
            )
            + normalize_transform
        )

        return instance_sampler

    def create_transformation(self) -> Transformation:
        """
        Add a transformation that replaces NaN in the input data with zeros,
        and mention whether the data was a NaN or not in another field.

        Returns:
        --------
        transformation: Transformation
            The chain of transformations defined for TACTiS.
        """
        return Chain(
            [
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                ),
            ]
        )

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: nn.Module,
        device: torch.device,
        experiment_mode: str = "forecasting",
        history_length: int = -1,
    ) -> PyTorchPredictor:
        """
        Create the predictor which can be used by GluonTS to do inference.

        Parameters:
        -----------
        transformation: Transformation
            The transformation to apply to the data prior to being sent to the model.
        trained_network: nn.Module
            An instance of TACTiSTrainingNetwork with trained parameters.
        device: torch.device
            The device where the model parameters should be placed.

        Returns:
        --------
        predictor: PyTorchPredictor
            The PyTorchTS predictor object.
        """
        if experiment_mode == "forecasting":
            prediction_network = TACTiSPredictionNetwork(
                num_series=self.num_series,
                model_parameters=self.model_parameters,
                prediction_length=self.prediction_length,
                num_parallel_samples=self.num_parallel_samples,
            ).to(device=device)
        else:
            prediction_network = TACTiSPredictionNetworkInterpolation(
                num_series=self.num_series,
                model_parameters=self.model_parameters,
                prediction_length=self.prediction_length,
                history_length=history_length,
                num_parallel_samples=self.num_parallel_samples,
            ).to(device=device)
        copy_parameters(trained_network, prediction_network)

        output_transform = cdf_to_gaussian_forward_transform if self.cdf_normalization else None
        input_names = get_module_forward_input_names(prediction_network)
        prediction_splitter = self.create_instance_splitter("test")

        return PyTorchPredictor(
            input_transform=transformation + prediction_splitter,
            output_transform=output_transform,
            input_names=input_names,
            prediction_net=prediction_network,
            batch_size=self.trainer.batch_size,
            freq=self.freq,
            prediction_length=self.prediction_length,
            device=device,
        )

    def train_model(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        num_workers: int = 0,
        prefetch_factor: int = 2,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        optimizer: str = "adam",
        backtesting=False,
        **kwargs,
    ) -> TrainOutput:
        transformation = self.create_transformation()

        trained_net = self.create_training_network(self.trainer.device)

        input_names = get_module_forward_input_names(trained_net)

        with env._let(max_idle_transforms=maybe_len(training_data) or 0):
            training_instance_splitter = self.create_instance_splitter("training")

        training_iter_dataset = TransformedIterableDataset(
            dataset=training_data,
            transform=transformation + training_instance_splitter + SelectFields(input_names),
            is_train=True,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
        )

        training_data_loader = DataLoader(
            training_iter_dataset,
            batch_size=self.trainer.batch_size,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=True,
            worker_init_fn=self._worker_init_fn,
            **kwargs,
        )

        validation_instance_splitter = self.create_instance_splitter("validation")

        input_transform = transformation + validation_instance_splitter + SelectFields(input_names)
        if not backtesting:
            validation_iter_dataset = TransformedDataset(
                validation_data,
                transformation=SplitValidationTransform(self.history_length + self.prediction_length),
            )
        else:
            validation_iter_dataset = validation_data
        validation_iter_args = {
            "dataset": validation_iter_dataset,
            "transform": input_transform,
            "stack_fn": lambda data: batchify(data, self.trainer.device),
        }

        self.trainer(
            net=trained_net,
            train_iter=training_data_loader,
            validation_iter_args=validation_iter_args,
            optimizer=optimizer,
        )

        return trained_net

    def train(
        self,
        training_data: Dataset,
        validation_data: Optional[Dataset] = None,
        num_workers: int = 0,
        prefetch_factor: int = 2,
        shuffle_buffer_length: Optional[int] = None,
        cache_data: bool = False,
        backtesting: bool = False,
        **kwargs,
    ) -> PyTorchPredictor:
        train_model_output = self.train_model(
            training_data,
            validation_data,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            shuffle_buffer_length=shuffle_buffer_length,
            cache_data=cache_data,
            backtesting=backtesting,
            **kwargs,
        )
        return train_model_output

    def validate_model(
        self,
        validation_data: Optional[Dataset] = None,
        backtesting=False,
    ):
        transformation = self.create_transformation()

        trained_net = self.create_training_network(self.trainer.device)

        input_names = get_module_forward_input_names(trained_net)

        validation_instance_splitter = self.create_instance_splitter("validation")

        input_transform = transformation + validation_instance_splitter + SelectFields(input_names)
        if not backtesting:
            validation_iter_dataset = TransformedDataset(
                validation_data,
                transformation=SplitValidationTransform(self.history_length + self.prediction_length),
            )
        else:
            validation_iter_dataset = validation_data
        validation_iter_args = {
            "dataset": validation_iter_dataset,
            "transform": input_transform,
            "stack_fn": lambda data: batchify(data, self.trainer.device),
        }

        nll = self.trainer.validate(
            net=trained_net,
            validation_iter_args=validation_iter_args,
        )

        return nll

    def validate(
        self,
        validation_data: Optional[Dataset] = None,
        backtesting: bool = False,
    ):
        return self.validate_model(
            validation_data,
            backtesting=backtesting,
        )



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import torch


def check_memory(cuda_device):
    """Check the total memory and occupied memory for GPU"""
    devices_info = (
        os.popen('"/usr/bin/nvidia-smi" --query-gpu=memory.total,memory.used --format=csv,nounits,noheader')
        .read()
        .strip()
        .split("\n")
    )
    total, used = devices_info[int(cuda_device)].split(",")
    return total, used


def occupy_memory(cuda_device):
    """Create a large tensor and delete it.
    This operation occupies the GPU memory, so other processes cannot use the occupied memory.
    It is used to ensure that this process won't be stopped when it requires additional GPU memory.
    Be careful with this operation. It will influence other people when you are sharing GPUs with others.
    """
    total, used = check_memory(cuda_device)
    total = int(total)
    used = int(used)
    max_mem = int(total * 0.90)
    print("Total memory: " + str(total) + ", used memory: " + str(used))
    block_mem = max_mem - used
    if block_mem > 0:
        x = torch.cuda.FloatTensor(256, 1024, block_mem)
        del x


def set_gpu(x):
    """Set up which GPU we use for this process"""
    os.environ["CUDA_VISIBLE_DEVICES"] = x
    print("Using gpu:", x)



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Iterator, Tuple

import pandas as pd
from copy import deepcopy

from gluonts.dataset.common import DataEntry, Dataset
from gluonts.model.forecast import Forecast
from gluonts.model.predictor import Predictor
from gluonts.transform import AdhocTransform


def make_evaluation_predictions(
    dataset: Dataset, predictor: Predictor, num_samples: int = 100
) -> Tuple[Iterator[Forecast], Iterator[pd.Series]]:
    """
    Returns predictions for the trailing prediction_length observations of the given
    time series, using the given predictor.

    The predictor will take as input the given time series without the trailing
    prediction_length observations.

    Parameters
    ----------
    dataset
        Dataset where the evaluation will happen. Only the portion excluding
        the prediction_length portion is used when making prediction.
    predictor
        Model used to draw predictions.
    num_samples
        Number of samples to draw on the model when evaluating. Only sampling-based
        models will use this.

    Returns
    -------
    Tuple[Iterator[Forecast], Iterator[pd.Series]]
        A pair of iterators, the first one yielding the forecasts, and the second
        one yielding the corresponding ground truth series.
    """

    prediction_length = predictor.prediction_length
    freq = predictor.freq
    lead_time = predictor.lead_time

    def add_ts_dataframe(
        data_iterator: Iterator[DataEntry],
    ) -> Iterator[DataEntry]:
        for data_entry in data_iterator:
            data = data_entry.copy()
            index = pd.date_range(
                start=data["start"],
                freq=freq,
                periods=data["target"].shape[-1],
            )
            data["ts"] = pd.DataFrame(index=index, data=data["target"].transpose())
            yield data

    def ts_iter(dataset: Dataset) -> pd.DataFrame:
        for data_entry in add_ts_dataframe(iter(dataset)):
            yield data_entry["ts"]

    def truncate_target(data):
        data = data.copy()
        target = data["target"]
        assert target.shape[-1] >= prediction_length  # handles multivariate case (target_dim, history_length)
        data["target"] = target[..., : -prediction_length - lead_time]
        return data

    dataset_as_is = deepcopy(dataset)
    dataset_trunc = AdhocTransform(truncate_target).apply(dataset)

    return (
        predictor.predict(dataset_trunc, num_samples=num_samples),
        ts_iter(dataset_as_is),
    )



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
from typing import Optional, Union

import torch
import os
import gc
import torch.nn as nn
from torch.optim import Adam, RMSprop
from torch.utils.data import DataLoader

from gluonts.core.component import validated
from tactis.gluon.utils import (
    save_checkpoint,
    load_checkpoint,
    set_seed,
)
from pts import Trainer

from gluonts.dataset.loader import ValidationDataLoader


class TACTISTrainer(Trainer):
    @validated()
    def __init__(
        self,
        epochs: int = 100,
        epochs_phase_1=None,
        epochs_phase_2=None,
        batch_size: int = 32,
        training_num_batches_per_epoch: int = 50,
        learning_rate: float = 1e-3,
        weight_decay: float = 1e-6,
        clip_gradient: Optional[float] = None,
        device: Optional[Union[torch.device, str]] = None,
        log_subparams_every=-1,
        checkpoint_dir=None,
        seed=42,
        load_checkpoint=None,
        early_stopping_epochs=-1,
        do_not_restrict_time=False,
        skip_batch_size_search=False,
    ) -> None:
        """
        When `epochs_phase_1` and `epochs_phase_2` are not specified, training is done for a total of `epochs` epochs, and `early_stopping_epochs` is used to switch from phase 1 to phase 2.
        When `epochs_phase_1` and `epochs_phase_2` are specified, `early_stopping_epochs` is ignored, and phase 1 and phase 2 are trained for `epochs_phase_1` and `epochs_phase_2` epochs respectively.
        """
        self.epochs = epochs
        self.epochs_phase_1 = epochs_phase_1
        self.epochs_phase_2 = epochs_phase_2
        self.batch_size = batch_size
        self.training_num_batches_per_epoch = training_num_batches_per_epoch
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.clip_gradient = clip_gradient
        self.device = device
        self.log_subparams_every = log_subparams_every
        self.checkpoint_dir = checkpoint_dir
        self.seed = seed
        self.load_checkpoint = load_checkpoint

        self.early_stopping_epochs = early_stopping_epochs
        self.do_not_restrict_time = do_not_restrict_time
        self.skip_batch_size_search = skip_batch_size_search

        if self.checkpoint_dir == None:
            raise Exception("Checkpoint directory (checkpoint_dir) is required to be specified for training.")
        else:
            print("Checkpoints will be saved at", self.checkpoint_dir)
            os.makedirs(self.checkpoint_dir, exist_ok=True)

    def set_load_checkpoint(self, load_checkpoint):
        self.load_checkpoint = load_checkpoint

    @staticmethod
    def disable_grads(net, disable_grads):
        for j, p in enumerate(net.model.named_parameters()):
            for param in disable_grads:
                if param in p[0]:
                    p[1].requires_grad = False

    def initialize_stage_1(self, net, optimizer_name, ckpt=None):
        net.model.set_stage(1)
        if optimizer_name == "rmsprop":
            optim = RMSprop(
                net.parameters(),
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )
        elif optimizer_name == "adam":
            optim = Adam(
                net.parameters(),
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )

        if ckpt:
            print("Loading model...")
            net.load_state_dict(ckpt["model"])
            print("Loaded model")

            print("Loading optim...")
            optim.load_state_dict(ckpt["optim"])
            print("Loaded optim...")

        return net, optim

    def switch_to_stage_2(self, net, optimizer_name, ckpt=None):
        disable_grads = []

        net.model.set_stage(2)
        net.model.initialize_stage2()
        net.to(self.device)

        # Create parameter groups for the specific parameters
        # A list of parameters related to each component for reference
        params_input_encoder_flow = [
            "flow_series_encoder",
            "flow_time_encoding",
            "flow_input_encoder",
        ]
        params_encoder_flow = ["flow_encoder"]
        params_decoder_flow = ["decoder.marginal"]

        parameter_names_to_optimize = [
            "copula_series_encoder",
            "copula_time_encoding",
            "copula_input_encoder",
            "copula_encoder",
            "decoder.copula",
        ]
        params_to_optimize_in_stage2 = []
        for name, param in net.named_parameters():
            if any(pname in name for pname in parameter_names_to_optimize):
                params_to_optimize_in_stage2.append(param)

        if optimizer_name == "rmsprop":
            optim = RMSprop(
                params_to_optimize_in_stage2,
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )
        elif optimizer_name == "adam":
            optim = Adam(
                params_to_optimize_in_stage2,
                lr=self.learning_rate,
                weight_decay=self.weight_decay,
            )
        if ckpt:
            print("Loading model...")
            net.load_state_dict(ckpt["model"])
            print("Loaded model")

            print("Loading optim...")
            optim.load_state_dict(ckpt["optim"])
            print("Loaded optim...")

        disable_grads.extend(params_decoder_flow)
        disable_grads.extend(params_input_encoder_flow)
        disable_grads.extend(params_encoder_flow)
        TACTISTrainer.disable_grads(net, disable_grads)

        return net, optim

    def __call__(
        self,
        net: nn.Module,
        train_iter: DataLoader,
        validation_iter_args=None,
        optimizer: str = "adam",
    ) -> None:
        set_seed(self.seed)

        # Load checkpoint automatically detecting the stage
        if self.load_checkpoint:
            set_seed(self.seed)
            assert os.path.isfile(self.load_checkpoint), "Checkpoint " + self.load_checkpoint + "is invalid"
            print("Loading from checkpoint", self.load_checkpoint)
            ckpt = torch.load(self.load_checkpoint, map_location=self.device)
            if ckpt["stage"] == 1:
                print("Loaded checkpoint has stage 1")
                net.model.set_stage(1)
                flow_loss_weight = 1.0
                copula_loss_weight = 0.0
                net, optim = self.initialize_stage_1(net, optimizer, ckpt)
                current_stage = 1
            else:
                print("Loaded checkpoint has stage 2")
                flow_loss_weight = 0.0
                copula_loss_weight = 1.0
                net, optim = self.switch_to_stage_2(net, optimizer, ckpt)
                current_stage = 2
            start_epoch = ckpt["epoch"] + 1
            print("Start epoch set to", start_epoch)
            best_val_loss_unweighted = ckpt["best_val_loss_unweighted"]
            best_epoch = ckpt["best_epoch"]
            total_training_only_time = ckpt["total_training_only_time"]
        else:
            flow_loss_weight = 1.0
            copula_loss_weight = 0.0
            net, optim = self.initialize_stage_1(net, optimizer)
            current_stage = 1

        ### The following code contains a batch size search.
        ### It can be used to maximize the batch size that one can use with the available GPU memory.
        ### It is however completely optional and may be skipped using the "skip_batch_size_search" argument
        ### If skipped, the batch size can be manually set appropriately
        if not self.skip_batch_size_search:
            print("Performing a batch size search with 10 iterations.")
            net.train()
            batch_size = self.batch_size
            print("Initial batch size:", batch_size)
            iters_since_batch_size = 0
            for batch_no, data_entry in enumerate(train_iter, start=0):
                try:
                    inputs = [v.to(self.device)[:batch_size] for v in data_entry.values()]
                    _ = net(*inputs)
                    # Loss computation
                    marginal_logdet, copula_loss = (
                        net.model.marginal_logdet,
                        net.model.copula_loss,
                    )
                    loss = copula_loss_weight * copula_loss - flow_loss_weight * marginal_logdet
                    loss_avg = loss.mean()
                    loss_avg.backward()
                    iters_since_batch_size += 1
                except RuntimeError as e:
                    if "out of memory" in str(e):
                        gc.collect()
                        torch.cuda.empty_cache()
                        print("Out of memory error encountered. Current batch size is", batch_size)
                        if batch_size == 1:
                            print("Batch is already at the minimum. Cannot reduce further. Exiting...")
                            return None
                        else:
                            print("Caught OutOfMemoryError. Reducing batch size...")
                            batch_size //= 2
                            print("New batch size:", batch_size)
                            iters_since_batch_size = 0
                    else:
                        print(e)
                        exit(1)
                if iters_since_batch_size == 10:
                    break
            print("Using a batch size of:", batch_size)

        # Initialise variables
        step = -1
        best_epoch = -1
        best_val_loss_unweighted = None

        start_epoch = 0
        epoch_no = start_epoch
        batch_size = self.batch_size
        total_training_only_time = 0.0

        ## Iterator for the epochs
        switch_to_stage_2 = False
        if self.epochs_phase_1 and self.epochs_phase_2:
            total_epochs = self.epochs_phase_1 + self.epochs_phase_2
        else:
            total_epochs = self.epochs

        ## Log the start and total epochs
        print("Start epoch:", start_epoch)
        print("Total Epochs:", total_epochs)

        print("Flow loss weight:", flow_loss_weight)
        print("Copula loss weight:", copula_loss_weight)

        while epoch_no < total_epochs:
            # Stage switching
            if switch_to_stage_2:
                print("Switching to stage 2")
                flow_loss_weight = 0.0
                copula_loss_weight = 1.0
                net, optim = self.switch_to_stage_2(net, optimizer)
                # Reset epochs, time
                epochs_since_best_epoch = 0
                total_training_only_time = 0.0
                # Set current stage
                current_stage = 2
                # Reset best validation loss
                best_val_loss_unweighted = None
                # Reset flag
                switch_to_stage_2 = False

            net.train()

            print("Epoch:", epoch_no, "/", total_epochs)
            set_seed(epoch_no + self.seed)
            cumm_epoch_loss = 0.0
            cumm_epoch_loss_unweighted = 0.0
            cumm_epoch_loss_unnormalized = 0.0
            cumm_marginal_logdet = 0.0  # Normalized
            cumm_copula_loss = 0.0  # Normalized

            print("Training.")
            print("Total number of training batches:", self.training_num_batches_per_epoch)

            # training loop
            training_num_windows_seen = 0
            training_epoch_start_time = time.time()
            for batch_no, data_entry in enumerate(train_iter, start=0):
                step = epoch_no * self.training_num_batches_per_epoch + batch_no
                if self.log_subparams_every != -1 and step % self.log_subparams_every == 0:
                    print("Iter:", batch_no, "/", self.training_num_batches_per_epoch)

                optim.zero_grad()

                inputs = [v.to(self.device)[:batch_size] for v in data_entry.values()]
                training_num_windows_seen += len(inputs[0])

                # Forward pass
                _ = net(*inputs)
                # Loss computation
                marginal_logdet, copula_loss = (
                    net.model.marginal_logdet,
                    net.model.copula_loss,
                )
                loss = copula_loss_weight * copula_loss - flow_loss_weight * marginal_logdet
                loss_avg = loss.mean()
                # Backward pass
                loss_avg.backward()
                if self.clip_gradient is not None:
                    nn.utils.clip_grad_norm_(net.parameters(), self.clip_gradient)
                optim.step()

                cumm_epoch_loss += loss.sum()
                cumm_epoch_loss_unweighted += torch.sum(copula_loss - marginal_logdet).item()
                cumm_epoch_loss_unnormalized += torch.sum(
                    net.model.unnormalized_copula_loss - net.model.marginal_logdet
                ).item()
                cumm_marginal_logdet += -torch.sum(marginal_logdet).item()
                cumm_copula_loss += torch.sum(copula_loss).item()

                if self.training_num_batches_per_epoch == batch_no:
                    break

            # Accumulate training time
            training_epoch_end_time = time.time()
            total_training_only_time += training_epoch_end_time - training_epoch_start_time

            avg_epoch_loss_unweighted = cumm_epoch_loss_unweighted / training_num_windows_seen
            print("Epoch:", epoch_no, "Average training loss:", avg_epoch_loss_unweighted)

            ####### VALIDATION #########
            ####### VALIDATION #########
            ####### VALIDATION #########
            if validation_iter_args is not None:
                validation_iter = ValidationDataLoader(**validation_iter_args, batch_size=batch_size)
            validation_length = None
            net.eval()
            cumm_epoch_loss_val = 0.0
            cumm_epoch_loss_val_unweighted = 0.0
            cumm_epoch_loss_val_unnormalized = 0.0
            cumm_marginal_loss_val = 0.0
            cumm_copula_loss_val = 0.0
            set_seed(epoch_no + self.seed)

            validation_num_windows_seen = 0
            validation_num_batches_per_epoch = len(validation_iter)

            print("Validation.")
            print("Total number of validation batches:", validation_num_batches_per_epoch)

            for batch_no, data_entry in enumerate(validation_iter, start=0):
                if self.log_subparams_every != -1 and step % self.log_subparams_every == 0:
                    print("Iter:", batch_no, "/", validation_num_batches_per_epoch)
                inputs = [v.to(self.device) for v in data_entry.values()]
                validation_num_windows_seen += len(inputs[0])
                if validation_length:
                    inputs = [v[:, :validation_length] for v in inputs]
                with torch.no_grad():
                    _ = net(*inputs)

                    marginal_logdet, copula_loss = (
                        net.model.marginal_logdet,
                        net.model.copula_loss,
                    )
                    loss = copula_loss_weight * copula_loss - flow_loss_weight * marginal_logdet
                    loss_avg = loss.mean()

                cumm_epoch_loss_val += loss.sum()
                cumm_epoch_loss_val_unweighted += torch.sum(copula_loss - marginal_logdet).item()
                cumm_epoch_loss_val_unnormalized += torch.sum(
                    net.model.unnormalized_copula_loss - net.model.marginal_logdet
                ).item()
                cumm_marginal_loss_val += -torch.sum(marginal_logdet).item()
                cumm_copula_loss_val += torch.sum(copula_loss).item()

            avg_epoch_loss_val_unweighted = cumm_epoch_loss_val_unweighted / validation_num_windows_seen

            print("Epoch:", epoch_no, "Average validation loss:", avg_epoch_loss_val_unweighted)

            if best_val_loss_unweighted == None or avg_epoch_loss_val_unweighted < best_val_loss_unweighted:
                best_val_loss_unweighted = avg_epoch_loss_val_unweighted
                best_epoch = epoch_no
                epochs_since_best_epoch = 0
            else:
                epochs_since_best_epoch += 1
            print("Epochs since best epoch:", epochs_since_best_epoch)

            ####################################################
            # All three stage-switching/early stopping criteria are below #
            ####################################################

            # Early stopping based shifting/stopping
            if (
                self.epochs_phase_1 is None
                and self.epochs_phase_2 is None
                and self.early_stopping_epochs != -1
                and epochs_since_best_epoch == self.early_stopping_epochs
            ):
                if net.model.stage == 2:
                    print("Stopping criterion reached for stage 2. Stopping training.")
                    # Load the best model
                    print("Loading the best model so far.")
                    checkpoint_path = os.path.join(self.checkpoint_dir, "best_stage_2.pth.tar")
                    net = load_checkpoint(checkpoint_path, net, self.device)
                    # Break
                    break
                else:
                    print("Stopping criterion reached for stage 1. Shifting to stage 2.")
                    # Load the best model
                    print("Loading the best model so far.")
                    checkpoint_path = os.path.join(self.checkpoint_dir, "best_stage_1.pth.tar")
                    net = load_checkpoint(checkpoint_path, net, self.device)
                    # Set flag
                    switch_to_stage_2 = True
                    # Increase epoch_no here since we are continuing
                    epoch_no += 1
                    # Continue
                    continue

            # Epochs based shifting/stopping
            if (
                net.model.stage == 2
                and self.epochs_phase_2
                and epoch_no - self.epochs_phase_1 == self.epochs_phase_2 - 1
            ):
                print(self.epochs_phase_2, "epochs completed in stage 2. Stopping training.")
                # Load the best model
                print("Loading the best model so far.")
                checkpoint_path = os.path.join(self.checkpoint_dir, "best_stage_2.pth.tar")
                net = load_checkpoint(checkpoint_path, net, self.device)
                # Break
                break
            elif net.model.stage == 1 and self.epochs_phase_1 and epoch_no == self.epochs_phase_1 - 1:
                print(self.epochs_phase_1, "epochs completed in stage 1. Shifting to stage 2.")
                # Load the best model
                print("Loading the best model so far.")
                checkpoint_path = os.path.join(self.checkpoint_dir, "best_stage_1.pth.tar")
                net = load_checkpoint(checkpoint_path, net, self.device)
                # Set flag
                switch_to_stage_2 = True
                # Increase epoch_no here since we are continuing
                epoch_no += 1
                # Continue
                continue

            # Training time based shifting / stopping
            if not self.do_not_restrict_time and total_training_only_time >= 129600:
                if net.model.stage == 2:
                    print("Time limit reached for stage 2. Stopping training.")
                    # Load the best model
                    print("Loading the best model so far.")
                    checkpoint_path = os.path.join(self.checkpoint_dir, "best_stage_2.pth.tar")
                    net = load_checkpoint(checkpoint_path, net, self.device)
                    # Break
                    break
                else:
                    print("Time limit reached for stage 1. Shifting to stage 2.")
                    # Load the best model
                    print("Loading the best model so far.")
                    checkpoint_path = os.path.join(self.checkpoint_dir, "best_stage_1.pth.tar")
                    net = load_checkpoint(checkpoint_path, net, self.device)
                    # Set flag
                    switch_to_stage_2 = True
                    # Increase epoch_no here since we are continuing
                    epoch_no += 1
                    # Continue
                    continue

            ####### VALIDATION #########
            ####### VALIDATION #########
            ####### VALIDATION #########

            state_dict = {
                "model": net.state_dict(),
                "epoch": epoch_no,
                "step": step,
                "best_val_loss_unweighted": best_val_loss_unweighted,
                "best_epoch": best_epoch,
                "stage": net.model.stage,
                "total_training_only_time": total_training_only_time,
            }
            state_dict["optim"] = optim.state_dict()
            filename = "last_stage_" + str(current_stage) + ".pth.tar"
            save_checkpoint(state_dict, self.checkpoint_dir, filename=filename)
            print(
                "Checkpoint of epoch",
                epoch_no,
                "saved at",
                os.path.join(self.checkpoint_dir, filename),
            )

            if epochs_since_best_epoch == 0:
                filename = "best_stage_" + str(current_stage) + ".pth.tar"
                save_checkpoint(state_dict, self.checkpoint_dir, filename=filename)
                print(
                    "Checkpoint of epoch",
                    epoch_no,
                    "saved at",
                    os.path.join(self.checkpoint_dir, filename),
                )

            # Increase epoch_no
            epoch_no += 1

            # Print a new line for better readability
            print("\n")

        return best_val_loss_unweighted

    def validate(
        self,
        net: nn.Module,
        validation_iter_args=None,
    ):
        if self.load_checkpoint:
            assert os.path.isfile(self.load_checkpoint), "Checkpoint " + self.load_checkpoint + "is invalid"
            print("Loading from checkpoint", self.load_checkpoint)
            ckpt = torch.load(self.load_checkpoint, map_location=self.device)
            net.load_state_dict(ckpt["model"])

        # Note that the NLL you obtain from this function is only comparable to TACTiS-style architecture
        # To make it comparable to non-TACTiS style architectures, you need to denormalize the loss of every sample/batch which is not currently implemented
        print("Validation...")
        batch_size = self.batch_size

        if validation_iter_args is not None:
            print("Creating a validation dataloader with a batch size of", batch_size)
            validation_iter = ValidationDataLoader(**validation_iter_args, batch_size=batch_size)
        validation_length = None
        net.eval()
        cumm_epoch_loss_val = 0.0

        validation_num_windows_seen = 0
        for batch_no, data_entry in enumerate(validation_iter, start=0):
            inputs = [v.to(self.device) for v in data_entry.values()]
            validation_num_windows_seen += len(inputs[0])
            if validation_length:
                inputs = [v[:, :validation_length] for v in inputs]
            with torch.no_grad():
                _ = net(*inputs)

                marginal_logdet, copula_loss = (
                    net.model.marginal_logdet,
                    net.model.copula_loss,
                )
                loss = copula_loss - marginal_logdet

            cumm_epoch_loss_val += loss.sum()

        avg_epoch_loss_val = cumm_epoch_loss_val / validation_num_windows_seen

        return avg_epoch_loss_val



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import os
import gc
import pickle
import sys
import torch
from typing import Dict, Iterable, Iterator, Optional

import numpy as np
import pandas as pd
from gluonts import transform
from gluonts.dataset.common import DataEntry, Dataset
from gluonts.evaluation import MultivariateEvaluator
from gluonts.model.forecast import Forecast
from gluonts.model.predictor import Predictor

from tactis.gluon.backtest import make_evaluation_predictions


class SplitValidationTransform(transform.FlatMapTransformation):
    """
    Split a dataset to do validation tests ending at each possible time step.
    A time step is possible if the resulting series is at least as long as the window_length parameter.
    """

    def __init__(self, window_length: int):
        super().__init__()
        self.window_length = window_length
        self.num_windows_seen = 0

    def flatmap_transform(self, data: DataEntry, is_train: bool) -> Iterator[DataEntry]:
        full_length = data["target"].shape[-1]
        for end_point in range(self.window_length, full_length + 1):
            data_copy = data.copy()
            data_copy["target"] = data["target"][..., :end_point]
            self.num_windows_seen += 1
            yield data_copy


class SuppressOutput:
    """
    Context controller to remove any printing to standard output and standard error.
    Inspired from:
    https://stackoverflow.com/questions/8391411/how-to-block-calls-to-print
    """

    def __enter__(self):
        self._stdout_bkp = sys.stdout
        self._stderr_bkp = sys.stderr
        sys.stdout = sys.stderr = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._stdout_bkp
        sys.stderr = self._stderr_bkp


def _compute_energy_score(target_data: np.array, samples: np.array, num_samples: int, beta: float) -> np.float32:
    """
    Compute the unnormalized energy score for a single forecast.

    Parameters:
    -----------
    target_data: np.array [two dimensions]
        The ground truth.
    samples: np.array [number of samples, two additional dimensions]
        The samples from the forecasting method to be assessed.
    num_samples: int
        The number of samples from the forecast.
    beta: float
        The beta parameter for the energy score.

    Returns
    -------
    score: np.float32
        The energy score.
    """
    # The Frobenius norm of a matrix is equal to the Euclidean norm of its element:
    # the square root of the sum of the square of its elements
    norm = np.linalg.norm(samples - target_data[None, :, :], ord="fro", axis=(1, 2))
    first_term = (norm**beta).mean()

    # For the second term of the energy score, we need two independant realizations of the distributions.
    # So we do a sum ignoring the i == j terms (which would be zero anyway), and we normalize by the
    # number of pairs of samples we have summed over.
    s = np.float32(0)
    for i in range(num_samples - 1):
        for j in range(i + 1, num_samples):
            norm = np.linalg.norm(samples[i] - samples[j], ord="fro")
            s += norm**beta
    second_term = s / (num_samples * (num_samples - 1) / 2)

    return first_term - 0.5 * second_term


def compute_energy_score(
    targets: Iterable[pd.DataFrame], forecasts: Iterable[Forecast], beta: float = 1.0
) -> np.float32:
    """
    Compute the non-normalized energy score for a multivariate stochastic prediction from samples.

    Parameters:
    -----------
    targets: Iterable[pd.DataFrame]
        The observed values, containing both the history and the prediction windows.
        Each element is taken independantly, and the result averaged over them.
    dataset: Iterable[Forecast]
        An object containing multiple samples of the probabilistic forecasts.
        This iterable should have the exact same length as targets.
    beta: float, default to 1.
        The energy score parameter, must be between 0 and 2, exclusive.

    Returns:
    --------
    result: np.float32
        A dictionary containing the various metrics
    """
    assert 0 < beta < 2

    cumulative_score = np.float32(0)
    num_forecasts = 0
    for target, forecast in zip(targets, forecasts):
        # The targets should always end with the prediction window
        assert target.index[-forecast.prediction_length] == forecast.start_date
        target_data = target.iloc[-forecast.prediction_length :].to_numpy()
        samples = forecast.samples

        cumulative_score += _compute_energy_score(target_data, samples, forecast.num_samples, beta)
        num_forecasts += 1
    return cumulative_score / num_forecasts


def compute_validation_metrics(
    predictor: Predictor,
    dataset: Dataset,
    window_length: int,
    prediction_length: int,
    num_samples: int,
    split: bool = True,
    savedir: Optional[str] = None,
    return_forecasts_and_targets: bool = False,
    subset_series=None,
    skip_energy=True,
    n_quantiles=20,
):
    if split:
        split_dataset = transform.TransformedDataset(dataset, transformation=SplitValidationTransform(window_length))
    else:
        split_dataset = dataset

    while True:
        print("Using batch size:", predictor.batch_size)
        try:
            forecast_it, ts_it = make_evaluation_predictions(
                dataset=split_dataset, predictor=predictor, num_samples=num_samples
            )
            forecasts = list(forecast_it)
            targets = list(ts_it)
            break
        except (torch.cuda.OutOfMemoryError, RuntimeError) as error:
            print(error)
            if predictor.batch_size == 1:
                print("Batch is already at the minimum. Cannot reduce further. Exiting...")
                return None
            else:
                print("Caught OutOfMemoryError. Reducing batch size...")
                predictor.batch_size //= 2
                gc.collect()
                torch.cuda.empty_cache()

    if subset_series:
        targets = [target.iloc[:, subset_series] for target in targets]
        for target in targets:
            target.columns = list(range(len(subset_series)))

    # A raw dump of the forecasts and targets for post-hoc analysis if needed, in the experiment folder
    # Can be loaded with the simple script:
    if savedir:
        savefile = os.path.join(savedir, "forecasts_targets.pkl")
        with open(savefile, "wb") as f:
            pickle.dump((forecasts, targets), f)

    # The results are going to be meaningless if any NaN shows up in the results,
    # so catch them here
    num_nan = 0
    num_inf = 0
    for forecast in forecasts:
        num_nan += np.isnan(forecast.samples).sum()
        num_inf += np.isinf(forecast.samples).sum()
    if num_nan > 0 or num_inf > 0:
        if skip_energy:
            return {
                "CRPS": float("nan"),
                "ND": float("nan"),
                "NRMSE": float("nan"),
                "MSE": float("nan"),
                "CRPS-Sum": float("nan"),
                "ND-Sum": float("nan"),
                "NRMSE-Sum": float("nan"),
                "MSE-Sum": float("nan"),
                "num_nan": num_nan,
                "num_inf": num_inf,
            }
        else:
            return {
                "CRPS": float("nan"),
                "ND": float("nan"),
                "NRMSE": float("nan"),
                "MSE": float("nan"),
                "CRPS-Sum": float("nan"),
                "ND-Sum": float("nan"),
                "NRMSE-Sum": float("nan"),
                "MSE-Sum": float("nan"),
                "Energy": float("nan"),
                "num_nan": num_nan,
                "num_inf": num_inf,
            }

    # Evaluate the quality of the model
    evaluator = MultivariateEvaluator(
        quantiles=(np.arange(n_quantiles) / float(n_quantiles))[1:],
        target_agg_funcs={"sum": np.sum},
    )

    # The GluonTS evaluator is very noisy on the standard error, so suppress it.
    with SuppressOutput():
        agg_metric, ts_wise_metrics = evaluator(targets, forecasts)

    if skip_energy:
        metrics = {
            "CRPS": agg_metric.get("mean_wQuantileLoss", float("nan")),
            "ND": agg_metric.get("ND", float("nan")),
            "NRMSE": agg_metric.get("NRMSE", float("nan")),
            "MSE": agg_metric.get("MSE", float("nan")),
            "CRPS-Sum": agg_metric.get("m_sum_mean_wQuantileLoss", float("nan")),
            "ND-Sum": agg_metric.get("m_sum_ND", float("nan")),
            "NRMSE-Sum": agg_metric.get("m_sum_NRMSE", float("nan")),
            "MSE-Sum": agg_metric.get("m_sum_MSE", float("nan")),
            "num_nan": num_nan,
            "num_inf": num_inf,
        }
    else:
        metrics = {
            "CRPS": agg_metric.get("mean_wQuantileLoss", float("nan")),
            "ND": agg_metric.get("ND", float("nan")),
            "NRMSE": agg_metric.get("NRMSE", float("nan")),
            "MSE": agg_metric.get("MSE", float("nan")),
            "CRPS-Sum": agg_metric.get("m_sum_mean_wQuantileLoss", float("nan")),
            "ND-Sum": agg_metric.get("m_sum_ND", float("nan")),
            "NRMSE-Sum": agg_metric.get("m_sum_NRMSE", float("nan")),
            "MSE-Sum": agg_metric.get("m_sum_MSE", float("nan")),
            "Energy": compute_energy_score(targets, forecasts),
            "num_nan": num_nan,
            "num_inf": num_inf,
        }

    if return_forecasts_and_targets:
        return metrics, ts_wise_metrics, forecasts, targets
    else:
        return metrics, ts_wise_metrics


def compute_validation_metrics_interpolation(
    predictor: Predictor,
    dataset: Dataset,
    window_length: int,
    prediction_length: int,
    num_samples: int,
    split: bool = True,
    savedir: Optional[str] = None,
    return_forecasts_and_targets: bool = False,
    subset_series=None,
    skip_energy=True,
    n_quantiles=20,
):
    if split:
        split_dataset = transform.TransformedDataset(dataset, transformation=SplitValidationTransform(window_length))
    else:
        raise Exception("split=False is not support in compute_validation_metrics_interpolation")

    while True:
        print("Using batch size:", predictor.batch_size)
        try:
            forecast_it, ts_it = make_evaluation_predictions(
                dataset=split_dataset, predictor=predictor, num_samples=num_samples
            )
            forecasts = list(forecast_it)
            targets = list(ts_it)
            break
        except (torch.cuda.OutOfMemoryError, RuntimeError) as error:
            print(error)
            if predictor.batch_size == 1:
                print("Batch is already at the minimum. Cannot reduce further. Exiting...")
                return None
            else:
                print("Caught OutOfMemoryError. Reducing batch size...")
                predictor.batch_size //= 2
                gc.collect()
                torch.cuda.empty_cache()

    if subset_series:
        targets = [target.iloc[:, subset_series] for target in targets]
        for target in targets:
            target.columns = list(range(len(subset_series)))

    # Store the original targets (interpolation "before" + prediction window + interpolation "after") for plotting/visualization purposes
    full_targets = []
    # Restrict targets array to until the interpolated segment to make it look like a forecasting task to MultivariateEvaluator
    # Also collect the start_date of the forecasts; and replace the forecast start_dates simultaneously
    # interpolation_segment_targets is the targets
    interpolation_segment_targets = []
    num_timesteps_observed_on_each_side = (window_length - prediction_length) // 2
    interpolation_window_start = num_timesteps_observed_on_each_side
    interpolation_window_end = num_timesteps_observed_on_each_side + prediction_length
    end_ts = interpolation_window_end

    for k, target in enumerate(targets):
        # Calculate offset
        offset = len(target) - window_length
        # Obtain and store the actual target window for interpolation
        modified_target = target.iloc[interpolation_window_start + offset : end_ts + offset]
        interpolation_segment_targets.append(modified_target)
        # Modify the start date of the window to the right one, overriding the one set by GluonTS based on the forecasting window
        forecasts[k].start_date = target.index[interpolation_window_start + offset]

        ## Store the original targets
        full_targets.append(target.iloc[offset : end_ts + num_timesteps_observed_on_each_side + offset])

    # Set targets to the new targets array
    targets = interpolation_segment_targets

    # A raw dump of the forecasts and targets for post-hoc analysis if needed, in the experiment folder
    # Can be loaded with the simple script:
    if savedir:
        savefile = os.path.join(savedir, "interpolation_targets.pkl")
        with open(savefile, "wb") as f:
            pickle.dump((forecasts, targets), f)

    # The results are going to be meaningless if any NaN shows up in the results,
    # so catch them here
    num_nan = 0
    num_inf = 0
    for forecast in forecasts:
        num_nan += np.isnan(forecast.samples).sum()
        num_inf += np.isinf(forecast.samples).sum()
    if num_nan > 0 or num_inf > 0:
        if skip_energy:
            return {
                "CRPS": float("nan"),
                "ND": float("nan"),
                "NRMSE": float("nan"),
                "MSE": float("nan"),
                "CRPS-Sum": float("nan"),
                "ND-Sum": float("nan"),
                "NRMSE-Sum": float("nan"),
                "MSE-Sum": float("nan"),
                "num_nan": num_nan,
                "num_inf": num_inf,
            }
        else:
            return {
                "CRPS": float("nan"),
                "ND": float("nan"),
                "NRMSE": float("nan"),
                "MSE": float("nan"),
                "CRPS-Sum": float("nan"),
                "ND-Sum": float("nan"),
                "NRMSE-Sum": float("nan"),
                "MSE-Sum": float("nan"),
                "Energy": float("nan"),
                "num_nan": num_nan,
                "num_inf": num_inf,
            }

    # Evaluate the quality of the model
    evaluator = MultivariateEvaluator(
        quantiles=(np.arange(n_quantiles) / float(n_quantiles))[1:],
        target_agg_funcs={"sum": np.sum},
    )

    # The GluonTS evaluator is very noisy on the standard error, so suppress it.
    with SuppressOutput():
        agg_metric, ts_wise_metrics = evaluator(targets, forecasts)

    if skip_energy:
        metrics = {
            "CRPS": agg_metric.get("mean_wQuantileLoss", float("nan")),
            "ND": agg_metric.get("ND", float("nan")),
            "NRMSE": agg_metric.get("NRMSE", float("nan")),
            "MSE": agg_metric.get("MSE", float("nan")),
            "CRPS-Sum": agg_metric.get("m_sum_mean_wQuantileLoss", float("nan")),
            "ND-Sum": agg_metric.get("m_sum_ND", float("nan")),
            "NRMSE-Sum": agg_metric.get("m_sum_NRMSE", float("nan")),
            "MSE-Sum": agg_metric.get("m_sum_MSE", float("nan")),
            "num_nan": num_nan,
            "num_inf": num_inf,
        }
    else:
        metrics = {
            "CRPS": agg_metric.get("mean_wQuantileLoss", float("nan")),
            "ND": agg_metric.get("ND", float("nan")),
            "NRMSE": agg_metric.get("NRMSE", float("nan")),
            "MSE": agg_metric.get("MSE", float("nan")),
            "CRPS-Sum": agg_metric.get("m_sum_mean_wQuantileLoss", float("nan")),
            "ND-Sum": agg_metric.get("m_sum_ND", float("nan")),
            "NRMSE-Sum": agg_metric.get("m_sum_NRMSE", float("nan")),
            "MSE-Sum": agg_metric.get("m_sum_MSE", float("nan")),
            "Energy": compute_energy_score(targets, forecasts),
            "num_nan": num_nan,
            "num_inf": num_inf,
        }

    if return_forecasts_and_targets:
        # Note: we return the full targets since we need that for plotting
        return metrics, ts_wise_metrics, forecasts, full_targets
    else:
        return metrics, ts_wise_metrics



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from typing import Any, Dict, Optional, Tuple

import numpy
import torch
from copy import deepcopy
from torch import nn

from .decoder import CopulaDecoder
from .encoder import Encoder, TemporalEncoder


class PositionalEncoding(nn.Module):
    """
    A class implementing the positional encoding for Transformers described in Vaswani et al. (2017).
    Somewhat generalized to allow unaligned or unordered time steps, as long as the time steps are integers.

    Adapted from: https://pytorch.org/tutorials/beginner/transformer_tutorial.html
    """

    def __init__(self, embedding_dim: int, dropout: float = 0.1, max_length: int = 5000):
        """
        Parameters:
        -----------
        embedding_dim: int
            The dimension of the input and output embeddings for this encoding.
        dropout: float, default to 0.1
            Dropout parameter for this encoding.
        max_length: int, default to 5000
            The maximum time steps difference which will have to be handled by this encoding.
        """
        super().__init__()

        assert embedding_dim % 2 == 0, "PositionEncoding needs an even embedding dimension"

        self.dropout = nn.Dropout(p=dropout)

        pos_encoding = torch.zeros(max_length, embedding_dim)
        possible_pos = torch.arange(0, max_length, dtype=torch.float)[:, None]
        factor = torch.exp(torch.arange(0, embedding_dim, 2, dtype=torch.float) * (-numpy.log(10000.0) / embedding_dim))

        # Alternate between using sine and cosine
        pos_encoding[:, 0::2] = torch.sin(possible_pos * factor)
        pos_encoding[:, 1::2] = torch.cos(possible_pos * factor)

        # Register as a buffer, to automatically be sent to another device if the model is sent there
        self.register_buffer("pos_encoding", pos_encoding)

    def forward(self, input_encoded: torch.Tensor, timesteps: torch.IntTensor) -> torch.Tensor:
        """
        Parameters:
        -----------
        input_encoded: torch.Tensor [batch, series, time steps, embedding dimension]
            An embedding which will be modified by the position encoding.
        timesteps: torch.IntTensor [batch, series, time steps] or [batch, 1, time steps]
            The time step for each entry in the input.

        Returns:
        --------
        output_encoded: torch.Tensor [batch, series, time steps, embedding dimension]
            The modified embedding.
        """
        # Use the time difference between the first time step of each batch and the other time steps.
        # min returns two outputs, we only keep the first.
        min_t = timesteps.min(dim=1, keepdim=True)[0].min(dim=2, keepdim=True)[0]
        delta_t = timesteps - min_t

        output_encoded = input_encoded + self.pos_encoding[delta_t]
        return self.dropout(output_encoded)

    def get_pos_encoding(self, timesteps: torch.IntTensor) -> torch.Tensor:
        min_t = timesteps.min(dim=1, keepdim=True)[0].min(dim=2, keepdim=True)[0]
        delta_t = timesteps - min_t

        return self.pos_encoding[delta_t]


class NormalizationIdentity:
    """
    Trivial normalization helper. Do nothing to its data.
    """

    def __init__(self, hist_value: torch.Tensor):
        """
        Parameters:
        -----------
        hist_value: torch.Tensor [batch, series, time steps]
            Historical data which can be used in the normalization.
        """
        pass

    def normalize(self, value: torch.Tensor) -> torch.Tensor:
        """
        Normalize the given values according to the historical data sent in the constructor.

        Parameters:
        -----------
        value: Tensor [batch, series, time steps]
            A tensor containing the values to be normalized.

        Returns:
        --------
        norm_value: Tensor [batch, series, time steps]
            The normalized values.
        """
        return value

    def denormalize(self, norm_value: torch.Tensor) -> torch.Tensor:
        """
        Undo the normalization done in the normalize() function.

        Parameters:
        -----------
        norm_value: Tensor [batch, series, time steps, samples]
            A tensor containing the normalized values to be denormalized.

        Returns:
        --------
        value: Tensor [batch, series, time steps, samples]
            The denormalized values.
        """
        return norm_value


class NormalizationStandardization:
    """
    Normalization helper for the standardization.

    The data for each batch and each series will be normalized by:
    - substracting the historical data mean,
    - and dividing by the historical data standard deviation.

    Use a lower bound of 1e-8 for the standard deviation to avoid numerical problems.
    """

    def __init__(self, hist_value: torch.Tensor):
        """
        Parameters:
        -----------
        hist_value: torch.Tensor [batch, series, time steps]
            Historical data which can be used in the normalization.
        """
        std, mean = torch.std_mean(hist_value, dim=2, unbiased=True, keepdim=True)
        self.std = std.clamp(min=1e-8)
        self.mean = mean

    def normalize(self, value: torch.Tensor) -> torch.Tensor:
        """
        Normalize the given values according to the historical data sent in the constructor.

        Parameters:
        -----------
        value: Tensor [batch, series, time steps]
            A tensor containing the values to be normalized.

        Returns:
        --------
        norm_value: Tensor [batch, series, time steps]
            The normalized values.
        """
        value = (value - self.mean) / self.std
        return value

    def denormalize(self, norm_value: torch.Tensor) -> torch.Tensor:
        """
        Undo the normalization done in the normalize() function.

        Parameters:
        -----------
        norm_value: Tensor [batch, series, time steps, samples]
            A tensor containing the normalized values to be denormalized.

        Returns:
        --------
        value: Tensor [batch, series, time steps, samples]
            The denormalized values.
        """
        norm_value = (norm_value * self.std[:, :, :, None]) + self.mean[:, :, :, None]
        return norm_value


class TACTiS(nn.Module):
    """
    The top-level module for TACTiS.

    The role of this module is to handle everything outside of the encoder and decoder.
    This consists mainly the data manipulation ahead of the encoder and after the decoder.
    """

    def __init__(
        self,
        num_series: int,
        flow_series_embedding_dim: int,
        copula_series_embedding_dim: int,
        flow_input_encoder_layers: int,
        copula_input_encoder_layers: int,
        bagging_size: Optional[int] = None,
        input_encoding_normalization: bool = True,
        data_normalization: str = "none",
        loss_normalization: str = "series",
        positional_encoding: Optional[Dict[str, Any]] = None,
        flow_encoder: Optional[Dict[str, Any]] = None,
        copula_encoder: Optional[Dict[str, Any]] = None,
        flow_temporal_encoder: Optional[Dict[str, Any]] = None,
        copula_temporal_encoder: Optional[Dict[str, Any]] = None,
        copula_decoder: Optional[Dict[str, Any]] = None,
        skip_copula: bool = True,
        experiment_mode: str = "forecasting",
    ):
        """
        Parameters:
        -----------
        num_series: int
            Number of series of the data which will be sent to the model. Required for the series embedding.
        bagging_size: Optional[int], default to None
            If set, the loss() method will only consider a random subset of the series at each call.
            The number of series kept is the value of this parameter.
        subset_series: default to None
            If set, should be a list of series to consider during training/evaluation. Used for toy experiments but with large datasets.
            Most importantly, different from bagging, here, the series is also used at inference.
        input_encoding_normalization: bool, default to True
            If true, the encoded input values (prior to the positional encoding) are scaled
            by the square root of their dimensionality.
        data_normalization: str ["", "none", "standardization"], default to "series"
            How to normalize the input values before sending them to the model.
        loss_normalization: str ["", "none", "series", "timesteps", "both"], default to "series"
            Scale the loss function by the number of series, timesteps, or both.
        positional_encoding: Optional[Dict[str, Any]], default to None
            If set to a non-None value, uses a PositionalEncoding for the time encoding.
            The options sent to the PositionalEncoding is content of this dictionary.
        copula_decoder: Optional[Dict[str, Any]], default to None
            If set to a non-None value, uses a CopulaDecoder as the decoder.
            The options sent to the CopulaDecoder is content of this dictionary.
        """
        super().__init__()

        assert copula_decoder is not None, "Must select exactly one type of decoder"

        assert (not bagging_size) or bagging_size <= num_series, "Bagging size must not be above number of series"

        data_normalization = data_normalization.lower()
        assert data_normalization in {"", "none", "standardization"}
        loss_normalization = loss_normalization.lower()
        assert loss_normalization in {"", "none", "series", "timesteps", "both"}

        self.num_series = num_series
        self.flow_series_embedding_dim = flow_series_embedding_dim
        self.copula_series_embedding_dim = copula_series_embedding_dim
        self.flow_input_encoder_layers = flow_input_encoder_layers
        self.copula_input_encoder_layers = copula_input_encoder_layers
        self.bagging_size = bagging_size
        self.input_encoding_normalization = input_encoding_normalization
        self.loss_normalization = loss_normalization
        self.positional_encoding = positional_encoding
        self.flow_encoder_args = flow_encoder
        self.copula_encoder_args = copula_encoder
        self.flow_temporal_encoder_args = flow_temporal_encoder
        self.copula_temporal_encoder_args = copula_temporal_encoder
        self.copula_decoder_args = copula_decoder

        self.skip_copula = skip_copula

        assert experiment_mode in ["forecasting", "interpolation"]
        self.experiment_mode = experiment_mode

        # Data normalization
        self.data_normalization = {
            "": NormalizationIdentity,
            "none": NormalizationIdentity,
            "standardization": NormalizationStandardization,
        }[data_normalization]

        # Actual encoder
        # Split encoder
        if flow_encoder is not None and copula_encoder is not None:
            self.flow_encoder = Encoder(**flow_encoder)
            if not self.skip_copula:
                self.copula_encoder = Encoder(**copula_encoder)
        elif flow_temporal_encoder is not None and copula_temporal_encoder is not None:
            self.flow_encoder = TemporalEncoder(**flow_temporal_encoder)
            if not self.skip_copula:
                self.copula_encoder = TemporalEncoder(**copula_temporal_encoder)
        self.flow_encoder_embedding_dim = self.flow_encoder.embedding_dim
        if not self.skip_copula:
            self.copula_encoder_embedding_dim = self.copula_encoder.embedding_dim
            copula_decoder["attentional_copula"]["input_dim"] = self.copula_encoder_embedding_dim

        # Split input encoder
        # Series encoding, Positional encoding and input encoder (that transforms [x*m, c, m] into [z])
        self.flow_series_encoder = nn.Embedding(num_embeddings=num_series, embedding_dim=self.flow_series_embedding_dim)
        if not self.skip_copula:
            self.copula_series_encoder = nn.Embedding(
                num_embeddings=num_series,
                embedding_dim=self.copula_series_embedding_dim,
            )

        if positional_encoding is not None:
            self.flow_time_encoding = PositionalEncoding(self.flow_encoder_embedding_dim, **positional_encoding)
            if not self.skip_copula:
                self.copula_time_encoding = PositionalEncoding(self.copula_encoder_embedding_dim, **positional_encoding)
        else:
            self.flow_time_encoding = None
            if not self.skip_copula:
                self.copula_time_encoding = None

        flow_elayers = nn.ModuleList([])
        for i in range(self.flow_input_encoder_layers):
            if i == 0:
                flow_elayers.append(nn.Linear(self.flow_series_embedding_dim + 2, self.flow_encoder_embedding_dim))
            else:
                flow_elayers.append(nn.Linear(self.flow_encoder_embedding_dim, self.flow_encoder_embedding_dim))
            flow_elayers.append(nn.ReLU())
        self.flow_input_encoder = nn.Sequential(*flow_elayers)

        if not self.skip_copula:
            copula_elayers = nn.ModuleList([])
            for i in range(self.copula_input_encoder_layers):
                if i == 0:
                    copula_elayers.append(
                        nn.Linear(self.copula_series_embedding_dim + 2, self.copula_encoder_embedding_dim)
                    )  # +1 for the value, +1 for the mask, and the per series embedding
                else:
                    copula_elayers.append(
                        nn.Linear(self.copula_encoder_embedding_dim, self.copula_encoder_embedding_dim)
                    )
                copula_elayers.append(nn.ReLU())
            self.copula_input_encoder = deepcopy(nn.Sequential(*copula_elayers))

        if copula_decoder is not None:
            flow_input_dim = self.flow_encoder_embedding_dim
            copula_input_dim = (
                None if self.skip_copula else self.copula_encoder_embedding_dim
            )  # Since we do not have a copula encoder yet
            self.decoder = CopulaDecoder(
                flow_input_dim=flow_input_dim,
                copula_input_dim=copula_input_dim,
                skip_copula=self.skip_copula,
                **copula_decoder
            )

        self.stage = 1
        self.copula_loss = None
        self.marginal_logdet = None
        self.current_normalizer = None

    def set_stage(self, stage: int):
        self.stage = stage

    def initialize_stage2(self):
        self.set_stage(2)
        self.skip_copula = False
        if self.copula_encoder_args:
            self.copula_encoder = Encoder(**self.copula_encoder_args)
            self.decoder.attentional_copula_args["input_dim"] = self.copula_encoder.embedding_dim
        elif self.copula_temporal_encoder_args:
            self.copula_encoder = TemporalEncoder(**self.copula_temporal_encoder_args)
            self.decoder.attentional_copula_args["input_dim"] = self.copula_encoder.embedding_dim
        self.copula_encoder_embedding_dim = self.copula_encoder.embedding_dim
        copula_dim = self.copula_encoder_embedding_dim
        self.copula_series_encoder = nn.Embedding(
            num_embeddings=self.num_series,
            embedding_dim=self.copula_series_embedding_dim,
        )
        if self.positional_encoding:
            self.copula_time_encoding = PositionalEncoding(copula_dim, **self.positional_encoding)

        copula_elayers = nn.ModuleList([])
        for i in range(self.copula_input_encoder_layers):
            if i == 0:
                copula_elayers.append(
                    nn.Linear(self.copula_series_embedding_dim + 2, copula_dim)
                )  # +1 for the value, +1 for the mask, and the per series embedding
            else:
                copula_elayers.append(nn.Linear(copula_dim, copula_dim))
            copula_elayers.append(nn.ReLU())
        self.copula_input_encoder = deepcopy(nn.Sequential(*copula_elayers))

        if self.copula_decoder_args:
            self.decoder.create_attentional_copula()

    def set_experiment_mode(self, experiment_mode: str):
        assert experiment_mode in ["forecasting", "interpolation"]
        self.experiment_mode = experiment_mode

    @staticmethod
    def _apply_bagging(
        bagging_size,
        hist_time: torch.Tensor,
        hist_value: torch.Tensor,
        pred_time: torch.Tensor,
        pred_value: torch.Tensor,
        flow_series_emb=None,
        copula_series_emb=None,
    ):
        """
        Only keep a small number of series for each of the input tensors.
        Which series will be kept is randomly selected for each batch. The order is not preserved.

        Parameters:
        -----------
        bagging_size: int
            How many series to keep
        hist_time: Tensor [batch, series, time steps]
            A tensor containing the time steps associated with the values of hist_value.
        hist_value: Tensor [batch, series, time steps]
            A tensor containing the values that will be available at inference time.
        pred_time: Tensor [batch, series, time steps]
            A tensor containing the time steps associated with the values of pred_value.
        pred_value: Tensor [batch, series, time steps]
            A tensor containing the values that the model should learn to forecast at inference time.
        flow_series_emb: Tensor [batch, series, flow embedding size]
            An embedding for each series for the marginals, expanded over the batches.
        copula_series_emb: Tensor [batch, series, copula embedding size]
            An embedding for each series for the copula, expanded over the batches.

        Returns:
        --------
        hist_time: Tensor [batch, series, time steps]
        hist_value: Tensor [batch, series, time steps]
        pred_time: Tensor [batch, series, time steps]
        pred_value: Tensor [batch, series, time steps]
        flow_series_emb: Tensor [batch, series, flow embedding size]
        copula_series_emb: Tensor [batch, series, copula embedding size]
        """
        num_batches = hist_time.shape[0]
        num_series = hist_time.shape[1]

        # Make sure to have the exact same bag for all series
        bags = [torch.randperm(num_series)[0:bagging_size] for _ in range(num_batches)]

        hist_time = torch.stack([hist_time[i, bags[i], :] for i in range(num_batches)], dim=0)
        hist_value = torch.stack([hist_value[i, bags[i], :] for i in range(num_batches)], dim=0)
        pred_time = torch.stack([pred_time[i, bags[i], :] for i in range(num_batches)], dim=0)
        pred_value = torch.stack([pred_value[i, bags[i], :] for i in range(num_batches)], dim=0)

        flow_series_emb = torch.stack([flow_series_emb[i, bags[i], :] for i in range(num_batches)], dim=0)
        if type(copula_series_emb) != type(None):
            copula_series_emb = torch.stack([copula_series_emb[i, bags[i], :] for i in range(num_batches)], dim=0)

        return (
            hist_time,
            hist_value,
            pred_time,
            pred_value,
            flow_series_emb,
            copula_series_emb,
        )

    @staticmethod
    def _apply_subsetting(
        series_to_keep,
        hist_time: torch.Tensor,
        hist_value: torch.Tensor,
        pred_time: torch.Tensor,
        pred_value=None,
        permute_series=False,
        flow_series_emb: torch.Tensor = None,
        copula_series_emb: torch.Tensor = None,
    ):
        """
        Only keep a small number of series for each of the input tensors.
        Which series will be kept is randomly selected for each batch. The order is not preserved.

        Parameters:
        -----------
        series_to_keep: int
            How many series to keep
        hist_time: Tensor [batch, series, time steps]
            A tensor containing the time steps associated with the values of hist_value.
        hist_value: Tensor [batch, series, time steps]
            A tensor containing the values that will be available at inference time.
        pred_time: Tensor [batch, series, time steps]
            A tensor containing the time steps associated with the values of pred_value.
        pred_value: Tensor [batch, series, time steps]
            A tensor containing the values that the model should learn to forecast at inference time.
        series_emb: Tensor [batch, series, embedding size]
            An embedding for each series, expanded over the batches.

        Returns:
        --------
        hist_time: Tensor [batch, series, time steps]
        hist_value: Tensor [batch, series, time steps]
        pred_time: Tensor [batch, series, time steps]
        pred_value: Tensor [batch, series, time steps]
        series_emb: Tensor [batch, series, embedding size]
            A subset of the input data, where only the given number of series is kept for each batch.
        """
        num_batches = hist_time.shape[0]

        if permute_series:
            bags = [
                torch.Tensor(series_to_keep)[torch.randperm(len(series_to_keep))].long() for _ in range(num_batches)
            ]
        else:
            bags = [torch.Tensor(series_to_keep).long() for _ in range(num_batches)]

        hist_time = torch.stack([hist_time[i, bags[i], :] for i in range(num_batches)], dim=0)
        hist_value = torch.stack([hist_value[i, bags[i], :] for i in range(num_batches)], dim=0)
        pred_time = torch.stack([pred_time[i, bags[i], :] for i in range(num_batches)], dim=0)

        flow_series_emb = torch.stack([flow_series_emb[i, bags[i], :] for i in range(num_batches)], dim=0)
        if type(copula_series_emb) != type(None):
            copula_series_emb = torch.stack([copula_series_emb[i, bags[i], :] for i in range(num_batches)], dim=0)

        if type(pred_value) != type(None):
            pred_value = torch.stack([pred_value[i, bags[i], :] for i in range(num_batches)], dim=0)
            return (
                hist_time,
                hist_value,
                pred_time,
                pred_value,
                flow_series_emb,
                copula_series_emb,
            )
        else:
            return hist_time, hist_value, pred_time, flow_series_emb, copula_series_emb

    def loss(
        self,
        hist_time: torch.Tensor,
        hist_value: torch.Tensor,
        pred_time: torch.Tensor,
        pred_value: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the loss function of the model.

        Parameters:
        -----------
        hist_time: Tensor [batch, series, time steps] or [batch, 1, time steps] or [batch, time steps]
            A tensor containing the time steps associated with the values of hist_value.
            If the series dimension is singleton or missing, then the time steps are taken as constant across all series.
        hist_value: Tensor [batch, series, time steps]
            A tensor containing the values that will be available at inference time.
        pred_time: Tensor [batch, series, time steps] or [batch, 1, time steps] or [batch, time steps]
            A tensor containing the time steps associated with the values of pred_value.
            If the series dimension is singleton or missing, then the time steps are taken as constant across all series.
        pred_value: Tensor [batch, series, time steps]
            A tensor containing the values that the model should learn to forecast at inference time.

        Returns:
        --------
        loss: torch.Tensor []
            The loss function of TACTiS, with lower values being better. Averaged over batches.
        """
        num_batches = hist_value.shape[0]
        num_series = hist_value.shape[1]
        num_hist_timesteps = hist_value.shape[2]
        num_pred_timesteps = pred_value.shape[2]
        device = hist_value.device

        copula_series_emb = None

        # Gets the embedding for each series [batch, series, embedding size]
        # Expand over batches to be compatible with the bagging procedure, which select different series for each batch
        flow_series_emb = self.flow_series_encoder(torch.arange(num_series, device=device))
        if not self.skip_copula:
            copula_series_emb = self.copula_series_encoder(torch.arange(num_series, device=device))
        flow_series_emb = flow_series_emb[None, :, :].expand(num_batches, -1, -1)
        if not self.skip_copula:
            copula_series_emb = copula_series_emb[None, :, :].expand(num_batches, -1, -1)

        # Make sure that both time tensors are in the correct format
        if len(hist_time.shape) == 2:
            hist_time = hist_time[:, None, :]
        if len(pred_time.shape) == 2:
            pred_time = pred_time[:, None, :]
        if hist_time.shape[1] == 1:
            hist_time = hist_time.expand(-1, num_series, -1)  # the shape will be [batch, series, num_hist_timesteps]
        if pred_time.shape[1] == 1:
            pred_time = pred_time.expand(-1, num_series, -1)  # the shape will be [batch, series, num_pred_timesteps]

        # If the setup is interpolation, fix the length on each side of the interpolated window
        if self.experiment_mode == "interpolation":
            total_observed_timesteps_on_each_side = num_hist_timesteps / 2
            if int(total_observed_timesteps_on_each_side) != total_observed_timesteps_on_each_side:
                raise Exception("Odd history lengths haven't been handled for now")
            total_observed_timesteps_on_each_side = int(total_observed_timesteps_on_each_side)

        # Perform bagging
        if self.bagging_size:
            (
                hist_time,
                hist_value,
                pred_time,
                pred_value,
                flow_series_emb,
                copula_series_emb,
            ) = self._apply_bagging(
                self.bagging_size,
                hist_time,
                hist_value,
                pred_time,
                pred_value,
                flow_series_emb=flow_series_emb,
                copula_series_emb=copula_series_emb,
            )
            num_series = self.bagging_size

        # The normalizer uses the same parameters for both historical and prediction values
        if self.experiment_mode == "forecasting":
            # The normalizer uses the history's stats
            normalizer = self.data_normalization(hist_value)
            hist_value = normalizer.normalize(hist_value)
            pred_value = normalizer.normalize(pred_value)
            true_value = torch.cat(
                [
                    hist_value,
                    pred_value,
                ],
                dim=2,
            )
            # The mask is at the end
            mask = torch.cat(
                [
                    torch.ones(
                        num_batches,
                        num_series,
                        num_hist_timesteps,
                        dtype=bool,
                        device=device,
                    ),
                    torch.zeros(
                        num_batches,
                        num_series,
                        num_pred_timesteps,
                        dtype=bool,
                        device=device,
                    ),
                ],
                dim=2,
            )
        elif self.experiment_mode == "interpolation":
            # The normalizer uses the stats of the available windows
            all_values = torch.cat(
                [
                    hist_value,
                    pred_value,
                ],
                dim=2,
            )
            observed_values = torch.cat(
                [
                    all_values[:, :, :total_observed_timesteps_on_each_side],
                    all_values[:, :, -total_observed_timesteps_on_each_side:],
                ],
                dim=2,
            )
            missing_values = all_values[
                :,
                :,
                total_observed_timesteps_on_each_side:-total_observed_timesteps_on_each_side,
            ]
            normalizer = self.data_normalization(observed_values)
            observed_values = normalizer.normalize(observed_values)
            missing_values = normalizer.normalize(missing_values)
            # true_value is used only in the decoder along with the mask; so creating it for interpolation is useful
            true_value = torch.cat(
                [
                    observed_values[:, :, :total_observed_timesteps_on_each_side],
                    missing_values,
                    observed_values[:, :, -total_observed_timesteps_on_each_side:],
                ],
                dim=2,
            )
            # The mask has to be such that the window of interpolation comes in the middle of the entire window
            mask = torch.cat(
                [
                    torch.ones(
                        num_batches,
                        num_series,
                        total_observed_timesteps_on_each_side,
                        dtype=bool,
                        device=device,
                    ),
                    torch.zeros(
                        num_batches,
                        num_series,
                        num_pred_timesteps,
                        dtype=bool,
                        device=device,
                    ),
                    torch.ones(
                        num_batches,
                        num_series,
                        total_observed_timesteps_on_each_side,
                        dtype=bool,
                        device=device,
                    ),
                ],
                dim=2,
            )
        self.current_normalizer = normalizer

        if self.experiment_mode == "forecasting":
            # Forecasting
            hist_encoded_flow = torch.cat(
                [
                    hist_value[:, :, :, None],
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps, -1),
                    torch.ones(num_batches, num_series, num_hist_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            if not self.skip_copula:
                hist_encoded_copula = torch.cat(
                    [
                        hist_value[:, :, :, None],
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps, -1),
                        torch.ones(
                            num_batches,
                            num_series,
                            num_hist_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            # For the prediction embedding, replace the values by zeros, since they won't be available during sampling
            pred_encoded_flow = torch.cat(
                [
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros
            if not self.skip_copula:
                pred_encoded_copula = torch.cat(
                    [
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros

            flow_encoded = torch.cat(
                [hist_encoded_flow, pred_encoded_flow], dim=2
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]
            if not self.skip_copula:
                copula_encoded = torch.cat(
                    [hist_encoded_copula, pred_encoded_copula], dim=2
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]
            flow_encoded = self.flow_input_encoder(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_input_encoder(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if self.input_encoding_normalization:
                flow_encoded = flow_encoded * self.flow_encoder_embedding_dim**0.5
                if not self.skip_copula:
                    copula_encoded = copula_encoded * self.copula_encoder_embedding_dim**0.5

            # Add the time encoding here after the input encoding to be compatible with how positional encoding is used.
            # Adjustments may be required for other ways to encode time.
            timesteps = torch.cat([hist_time, pred_time], dim=2)
            flow_encoded = self.flow_time_encoding(flow_encoded, timesteps.to(int))
            if not self.skip_copula:
                copula_encoded = self.copula_time_encoding(copula_encoded, timesteps.to(int))

            flow_encoded = self.flow_encoder.forward(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_encoder.forward(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
        else:
            # Interpolation
            hist_encoded_flow = torch.cat(
                [
                    observed_values[:, :, :, None],
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps, -1),
                    torch.ones(num_batches, num_series, num_hist_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            if not self.skip_copula:
                hist_encoded_copula = torch.cat(
                    [
                        observed_values[:, :, :, None],
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps, -1),
                        torch.ones(
                            num_batches,
                            num_series,
                            num_hist_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            # For the prediction embedding, replace the values by zeros, since they won't be available during sampling
            pred_encoded_flow = torch.cat(
                [
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros
            if not self.skip_copula:
                pred_encoded_copula = torch.cat(
                    [
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros

            flow_encoded = torch.cat(
                [hist_encoded_flow, pred_encoded_flow], dim=2
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]
            if not self.skip_copula:
                copula_encoded = torch.cat(
                    [hist_encoded_copula, pred_encoded_copula], dim=2
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]

            # Important step for interpolation - reorder the encoded vectors to reflect the order of the mask
            # This is because everything else (mask, true_value) is in the correct timesteps order
            flow_encoded = torch.cat(
                [
                    flow_encoded[:, :, :total_observed_timesteps_on_each_side],
                    flow_encoded[:, :, -num_pred_timesteps:],
                    flow_encoded[
                        :,
                        :,
                        total_observed_timesteps_on_each_side : total_observed_timesteps_on_each_side
                        + total_observed_timesteps_on_each_side,
                    ],
                ],
                dim=2,
            )
            if not self.skip_copula:
                copula_encoded = torch.cat(
                    [
                        copula_encoded[:, :, :total_observed_timesteps_on_each_side],
                        copula_encoded[:, :, -num_pred_timesteps:],
                        copula_encoded[
                            :,
                            :,
                            total_observed_timesteps_on_each_side : total_observed_timesteps_on_each_side
                            + total_observed_timesteps_on_each_side,
                        ],
                    ],
                    dim=2,
                )

            flow_encoded = self.flow_input_encoder(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_input_encoder(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if self.input_encoding_normalization:
                flow_encoded = flow_encoded * self.flow_encoder_embedding_dim**0.5
                if not self.skip_copula:
                    copula_encoded = copula_encoded * self.copula_encoder_embedding_dim**0.5
            # Add the time encoding here after the input encoding to be compatible with how positional encoding is used.
            # Adjustments may be required for other ways to encode time.
            timesteps = torch.cat([hist_time, pred_time], dim=2)
            flow_encoded = self.flow_time_encoding(flow_encoded, timesteps.to(int))
            if not self.skip_copula:
                copula_encoded = self.copula_time_encoding(copula_encoded, timesteps.to(int))

            flow_encoded = self.flow_encoder.forward(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_encoder.forward(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]

        if self.skip_copula:
            copula_encoded = None

        _ = self.decoder.loss(
            flow_encoded=flow_encoded,
            copula_encoded=copula_encoded,
            mask=mask,
            true_value=true_value,
        )  # previously returned loss here when loss coefficients weren't used

        self.copula_loss = self.decoder.copula_loss
        self.marginal_logdet = self.decoder.marginal_logdet
        self.unnormalized_copula_loss = torch.clone(self.decoder.copula_loss)
        self.unnormalized_marginal_logdet = torch.clone(self.decoder.marginal_logdet)

        if self.loss_normalization in {"series", "both"}:
            self.copula_loss = self.copula_loss / num_series
            self.marginal_logdet = self.marginal_logdet / num_series
        if self.loss_normalization in {"timesteps", "both"}:
            self.copula_loss = self.copula_loss / num_pred_timesteps
            self.marginal_logdet = self.marginal_logdet / num_pred_timesteps

        return self.marginal_logdet, self.copula_loss

    def sample(
        self,
        num_samples: int,
        hist_time: torch.Tensor,
        hist_value: torch.Tensor,
        pred_time: torch.Tensor,
    ) -> torch.Tensor:
        """
        Generate the given number of samples from the forecasted distribution.

        Parameters:
        -----------
        num_samples: int
            How many samples to generate, must be >= 1.
        hist_time: Tensor [batch, series, time steps] or [batch, 1, time steps] or [batch, time steps]
            A tensor containing the times associated with the values of hist_value.
            If the series dimension is singleton or missing, then the time steps are taken as constant across all series.
        hist_value: Tensor [batch, series, time steps]
            A tensor containing the available values
        pred_time: Tensor [batch, series, time steps] or [batch, 1, time steps] or [batch, time steps]
            A tensor containing the times at which we want forecasts.
            If the series dimension is singleton or missing, then the time steps are taken as constant across all series.
        Returns:
        --------
        samples: torch.Tensor [batch, series, time steps, samples]
            Samples from the forecasted distribution.
        """
        num_batches = hist_value.shape[0]
        num_series = hist_value.shape[1]
        num_hist_timesteps = hist_value.shape[2]
        num_pred_timesteps = pred_time.shape[-1]
        device = hist_value.device

        # Gets the embedding for each series [batch, series, embedding size]
        # Expand over batches to be compatible with the bagging procedure, which select different series for each batch
        copula_series_emb = None

        flow_series_emb = self.flow_series_encoder(torch.arange(num_series, device=device))
        if not self.skip_copula:
            copula_series_emb = self.copula_series_encoder(torch.arange(num_series, device=device))
        flow_series_emb = flow_series_emb[None, :, :].expand(num_batches, -1, -1)
        if not self.skip_copula:
            copula_series_emb = copula_series_emb[None, :, :].expand(num_batches, -1, -1)

        # Make sure that both time tensors are in the correct format
        if len(hist_time.shape) == 2:
            hist_time = hist_time[:, None, :]
        if len(pred_time.shape) == 2:
            pred_time = pred_time[:, None, :]
        if hist_time.shape[1] == 1:
            hist_time = hist_time.expand(-1, num_series, -1)
        if pred_time.shape[1] == 1:
            pred_time = pred_time.expand(-1, num_series, -1)

        # If the setup is interpolation, fix the length on each side of the interpolated window
        if self.experiment_mode == "interpolation":
            actual_history_length = num_hist_timesteps - num_pred_timesteps
            total_observed_timesteps_on_each_side = actual_history_length / 2
            if int(total_observed_timesteps_on_each_side) != total_observed_timesteps_on_each_side:
                raise Exception("Odd history lengths haven't been handled for now")
            total_observed_timesteps_on_each_side = int(total_observed_timesteps_on_each_side)

        # If we are performing interpolation, we need to modify hist_time and pred_time to reflect the mask
        # We assume pred_time is invalid here since hist_time contains the entire window
        # But hist_time and pred_time should not reflect the mask since by the time the positional encoding is added, the order itself is correct
        # So we only use the `prediction_length` or `num_pred_timesteps` from pred_time. We throw away pred_time itself
        # Also remove the prediction window from `hist_value`
        if self.experiment_mode == "interpolation":
            # actual_hist_time = torch.cat(
            #     [
            #         hist_time[:, :, :total_observed_timesteps_on_each_side],
            #         hist_time[:, :, -total_observed_timesteps_on_each_side:],
            #     ],
            #     dim=2,
            # )
            # actual_pred_time = hist_time[:, :, total_observed_timesteps_on_each_side:total_observed_timesteps_on_each_side+num_pred_timesteps]
            hist_time = hist_time[:, :, :-num_pred_timesteps]
            pred_time = hist_time[:, :, -num_pred_timesteps:]
            hist_value = torch.cat(
                [
                    hist_value[:, :, :total_observed_timesteps_on_each_side],
                    hist_value[:, :, -total_observed_timesteps_on_each_side:],
                ],
                dim=2,
            )
            observed_values = hist_value

        # The normalizer remembers its parameter to reverse it with the samples
        if self.experiment_mode == "forecasting":
            normalizer = self.data_normalization(hist_value)
            hist_value = normalizer.normalize(hist_value)
            # Make the mask and true_value
            mask = torch.cat(
                [
                    torch.ones(
                        num_batches,
                        num_series,
                        num_hist_timesteps,
                        dtype=bool,
                        device=device,
                    ),
                    torch.zeros(
                        num_batches,
                        num_series,
                        num_pred_timesteps,
                        dtype=bool,
                        device=device,
                    ),
                ],
                dim=2,
            )
            true_value = torch.cat(
                [
                    hist_value,
                    torch.zeros(num_batches, num_series, num_pred_timesteps, device=device),
                ],
                dim=2,
            )
        else:
            normalizer = self.data_normalization(observed_values)
            observed_values = normalizer.normalize(observed_values)
            # Make the mask and true_value
            mask = torch.cat(
                [
                    torch.ones(
                        num_batches,
                        num_series,
                        total_observed_timesteps_on_each_side,
                        dtype=bool,
                        device=device,
                    ),
                    torch.zeros(
                        num_batches,
                        num_series,
                        num_pred_timesteps,
                        dtype=bool,
                        device=device,
                    ),
                    torch.ones(
                        num_batches,
                        num_series,
                        total_observed_timesteps_on_each_side,
                        dtype=bool,
                        device=device,
                    ),
                ],
                dim=2,
            )
            # true_value is used only in the decoder along with the mask; so creating it for interpolation is useful
            true_value = torch.cat(
                [
                    observed_values[:, :, :total_observed_timesteps_on_each_side],
                    torch.zeros(num_batches, num_series, num_pred_timesteps, device=device),
                    observed_values[:, :, -total_observed_timesteps_on_each_side:],
                ],
                dim=2,
            )

        if self.experiment_mode == "forecasting":
            hist_encoded_flow = torch.cat(
                [
                    hist_value[:, :, :, None],
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps, -1),
                    torch.ones(num_batches, num_series, num_hist_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            if not self.skip_copula:
                hist_encoded_copula = torch.cat(
                    [
                        hist_value[:, :, :, None],
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps, -1),
                        torch.ones(
                            num_batches,
                            num_series,
                            num_hist_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            # For the prediction embedding, replace the values by zeros, since they won't be available during sampling
            pred_encoded_flow = torch.cat(
                [
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros
            if not self.skip_copula:
                pred_encoded_copula = torch.cat(
                    [
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros

            flow_encoded = torch.cat(
                [hist_encoded_flow, pred_encoded_flow], dim=2
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]
            if not self.skip_copula:
                copula_encoded = torch.cat(
                    [hist_encoded_copula, pred_encoded_copula], dim=2
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]
            flow_encoded = self.flow_input_encoder(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_input_encoder(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if self.input_encoding_normalization:
                flow_encoded = flow_encoded * self.flow_encoder_embedding_dim**0.5
                if not self.skip_copula:
                    copula_encoded = copula_encoded * self.copula_encoder_embedding_dim**0.5

            # Add the time encoding here after the input encoding to be compatible with how positional encoding is used.
            # Adjustments may be required for other ways to encode time.
            timesteps = torch.cat([hist_time, pred_time], dim=2)
            flow_encoded = self.flow_time_encoding(flow_encoded, timesteps.to(int))
            if not self.skip_copula:
                copula_encoded = self.copula_time_encoding(copula_encoded, timesteps.to(int))

            flow_encoded = self.flow_encoder.forward(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_encoder.forward(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
        else:
            # Interpolation
            hist_encoded_flow = torch.cat(
                [
                    observed_values[:, :, :, None],
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_hist_timesteps - num_pred_timesteps, -1),
                    torch.ones(
                        num_batches,
                        num_series,
                        num_hist_timesteps - num_pred_timesteps,
                        1,
                        device=device,
                    ),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            if not self.skip_copula:
                hist_encoded_copula = torch.cat(
                    [
                        observed_values[:, :, :, None],
                        copula_series_emb[:, :, None, :].expand(
                            num_batches, -1, num_hist_timesteps - num_pred_timesteps, -1
                        ),
                        torch.ones(
                            num_batches,
                            num_series,
                            num_hist_timesteps - num_pred_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]
            # For the prediction embedding, replace the values by zeros, since they won't be available during sampling
            pred_encoded_flow = torch.cat(
                [
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                    flow_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                    torch.zeros(num_batches, num_series, num_pred_timesteps, 1, device=device),
                ],
                dim=3,
            )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros
            if not self.skip_copula:
                pred_encoded_copula = torch.cat(
                    [
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                        copula_series_emb[:, :, None, :].expand(num_batches, -1, num_pred_timesteps, -1),
                        torch.zeros(
                            num_batches,
                            num_series,
                            num_pred_timesteps,
                            1,
                            device=device,
                        ),
                    ],
                    dim=3,
                )  # Shape: [batch, num_series, num_hist_timesteps, embedding_size+2]. Note that we are also leaving out the values as zeros

            flow_encoded = torch.cat(
                [hist_encoded_flow, pred_encoded_flow], dim=2
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]
            if not self.skip_copula:
                copula_encoded = torch.cat(
                    [hist_encoded_copula, pred_encoded_copula], dim=2
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, embedding_size+2]

            # Important step for interpolation - reorder the encoded vectors to reflect the order of the mask
            flow_encoded = torch.cat(
                [
                    flow_encoded[:, :, :total_observed_timesteps_on_each_side],
                    flow_encoded[:, :, -num_pred_timesteps:],
                    flow_encoded[
                        :,
                        :,
                        total_observed_timesteps_on_each_side : total_observed_timesteps_on_each_side
                        + total_observed_timesteps_on_each_side,
                    ],
                ],
                dim=2,
            )
            if not self.skip_copula:
                copula_encoded = torch.cat(
                    [
                        copula_encoded[:, :, :total_observed_timesteps_on_each_side],
                        copula_encoded[:, :, -num_pred_timesteps:],
                        copula_encoded[
                            :,
                            :,
                            total_observed_timesteps_on_each_side : total_observed_timesteps_on_each_side
                            + total_observed_timesteps_on_each_side,
                        ],
                    ],
                    dim=2,
                )

            flow_encoded = self.flow_input_encoder(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_input_encoder(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if self.input_encoding_normalization:
                flow_encoded = flow_encoded * self.flow_encoder_embedding_dim**0.5
                if not self.skip_copula:
                    copula_encoded = copula_encoded * self.copula_encoder_embedding_dim**0.5
            # Add the time encoding here after the input encoding to be compatible with how positional encoding is used.
            # Adjustments may be required for other ways to encode time.
            timesteps = torch.cat([hist_time, pred_time], dim=2)
            flow_encoded = self.flow_time_encoding(flow_encoded, timesteps.to(int))
            if not self.skip_copula:
                copula_encoded = self.copula_time_encoding(copula_encoded, timesteps.to(int))

            flow_encoded = self.flow_encoder.forward(
                flow_encoded
            )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]
            if not self.skip_copula:
                copula_encoded = self.copula_encoder.forward(
                    copula_encoded
                )  # Shape: [batch, num_series, num_hist_timesteps+num_pred_timesteps, encoder_size]

        if self.skip_copula:
            copula_encoded = None
        decoder_output = self.decoder.sample(
            num_samples,
            mask=mask,
            true_value=true_value,
            flow_encoded=flow_encoded,
            copula_encoded=copula_encoded,
        )

        samples = decoder_output
        samples = normalizer.denormalize(samples)
        return samples



"""
Copyright 2023 ServiceNow
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Tuple, Optional

import random
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns

from copy import deepcopy
from gluonts.model.forecast import Forecast


def plot_single_forecast(
    forecast: Forecast,
    target: pd.Series,
    axes: plt.Axes,
    locator: matplotlib.ticker.Locator,
) -> None:
    """
    Plot the forecast for a single series, on the given Axes object.

    Parameters:
    -----------
    forecast: Forecast
        The univariate forecast object generated through the GluonTS interface.
    target: pd.Series
        The ground truth for the series.
    axes: plt.Axes
        The Axes object on which to draw the forecast.
    locator: matplotlib.ticker.Locator
        An object defining how the horizontal ticks will be selected.
    """
    # Add the last point of historical data to the forecast, for a smoother transition.
    last_hist_index = target.index[-forecast.prediction_length - 1]
    last_hist_value = target[-forecast.prediction_length - 1]

    forecast_index = [last_hist_index] + [x for x in forecast.index]

    def quantile(q: float) -> np.ndarray:
        return np.append(last_hist_value, forecast.quantile(q))

    # Plot the forecasts with 3 different quantile ranges.
    axes.fill_between(
        forecast_index,
        quantile(0.05),
        quantile(0.95),
        facecolor=(0.75, 0.75, 1),
        interpolate=True,
        label="5%-95%",
        zorder=1,
    )
    axes.fill_between(
        forecast_index,
        quantile(0.1),
        quantile(0.9),
        facecolor=(0.25, 0.25, 1),
        interpolate=True,
        label="10%-90%",
        zorder=2,
    )
    axes.fill_between(
        forecast_index,
        quantile(0.25),
        quantile(0.75),
        facecolor=(0.0, 0.0, 0.75),
        interpolate=True,
        label="25%-75%",
        zorder=3,
    )

    # Plot the forecast median
    axes.plot(
        forecast_index,
        quantile(0.5),
        color=(0.5, 0.5, 0.5),
        linewidth=3,
        label="50%",
        zorder=4,
    )

    # Plot the ground truth with a white highlight
    target.plot(
        ax=axes,
        color="#FFFFFF",
        alpha=1,
        linewidth=2,
        zorder=5,
    )
    target.plot(
        ax=axes,
        color="#FC6A08",
        alpha=1,
        linewidth=1.5,
        label="Ground Truth",
        zorder=6,
    )

    # Highlight the forecasting regions
    axes.set_xlim(axes.get_xlim())
    xmin = forecast.index[0] - (forecast.index[1] - forecast.index[0]) / 2
    xmax = forecast.index[-1] + (forecast.index[-1] - forecast.index[0])  # Beyond the limit of the plot
    axes.axvspan(xmin=xmin, xmax=xmax, facecolor=(0.2, 0.5, 0.2), alpha=0.1)

    # Allows for nice time-based ticks
    axes.xaxis.set_major_locator(deepcopy(locator))
    axes.tick_params(
        axis="both",
        length=5,
        width=1,
        color=(0.7, 0.7, 0.7),
        left=True,
        bottom=True,
    )


def plot_four_forecasts(
    forecasts: List[Forecast],
    targets: List[pd.DataFrame],
    selection: List[Tuple[int, int]],
    tick_freq: str = "day",
    history_length: Optional[int] = None,
    savefile: Optional[str] = None,
) -> None:
    """
    Plot the forecast for four series, from potentially multiple forecasts.

    Parameters:
    -----------
    forecasts: List[Forecast]
        A list of multivariate forecasts generated through the GluonTS interface.
    targets: List[pd.DataFrame]
        A list of multivariate ground truths.
    selection: List[Tuple[int, int]]
        A list of 4 pairs of integers, to select which series to plot.
        The first element of the pairs selects which forecast to plot, while the second selects which series to plot.
    tick_freq: str, from "day", "6 hours", or "4 months end"
        The frequency of the horizontal tick marks.
    history_length: Optional[int], default to None
        If set, how much history to plot from the ground-truth (minimum = 1).
        If not set, default to the prediction length of the forecasts.
    """
    assert len(selection) == 4, "plot_four_forecasts() can only plot 4 series at once."

    # Seaborn nice default
    sns.set()
    sns.set_context("paper", font_scale=1)

    fig, axs = plt.subplots(2, 2, figsize=(10, 7))

    locator = {
        "day": matplotlib.dates.DayLocator(),
        "6 hours": matplotlib.dates.HourLocator(range(0, 24, 6)),
        "4 months end": matplotlib.dates.MonthLocator(range(1, 13, 4), bymonthday=30),
    }[tick_freq]

    counter = 0
    for forecast_num, series_num in selection:
        forecast = forecasts[forecast_num]
        target = targets[forecast_num]

        single_forecast = forecast.copy_dim(series_num)
        if history_length is None:
            window_length = 2 * single_forecast.prediction_length
        else:
            window_length = history_length + single_forecast.prediction_length
        single_target = target[series_num][-window_length:]
        plot_single_forecast(
            forecast=single_forecast,
            target=single_target,
            axes=axs.flat[counter],
            locator=locator,
        )
        counter += 1
    plt.subplots_adjust(hspace=0.3)

    # Reorder the legend, and have a unique one for all 4 plots.
    handles, labels = axs.flat[0].get_legend_handles_labels()
    order = [5, 0, 1, 2, 3]
    handles = [handles[o] for o in order]
    labels = [labels[o] for o in order]
    fig.legend(handles, labels, bbox_to_anchor=(0.91, 0.89), loc="upper left")
    fig.show()

    if savefile:
        fig.savefig(savefile, bbox_inches="tight", pad_inches=0)
